Index: include/cbfs.h
===================================================================
--- include/cbfs.h	(revision 0)
+++ include/cbfs.h	(revision 7)
@@ -0,0 +1,174 @@
+/*
+ * This file is part of the coreboot project.
+ *
+ * Copyright (C) 2008 Jordan Crouse <jordan@cosmicpenguin.net>
+ *
+ * This file is dual-licensed. You can choose between:
+ *   - The GNU GPL, version 2, as published by the Free Software Foundation
+ *   - The revised BSD license (without advertising clause)
+ *
+ * ---------------------------------------------------------------------------
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; version 2 of the License.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA, 02110-1301 USA
+ * ---------------------------------------------------------------------------
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notice, this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ * 3. The name of the author may not be used to endorse or promote products
+ *    derived from this software without specific prior written permission.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
+ * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+ * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
+ * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
+ * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
+ * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
+ * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
+ * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
+ * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
+ * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+ * SUCH DAMAGE.
+ * ---------------------------------------------------------------------------
+ */
+
+#ifndef _CBFS_H_
+#define _CBFS_H_
+#include <coreboot_tables.h>
+/** These are standard values for the known compression
+    alogrithms that coreboot knows about for stages and
+    payloads.  Of course, other LAR users can use whatever
+    values they want, as long as they understand them. */
+
+#define CBFS_COMPRESS_NONE  0
+#define CBFS_COMPRESS_LZMA  1
+
+/** These are standard component types for well known
+    components (i.e - those that coreboot needs to consume.
+    Users are welcome to use any other value for their
+    components */
+
+#define CBFS_TYPE_STAGE     0x10
+#define CBFS_TYPE_PAYLOAD   0x20
+#define CBFS_TYPE_OPTIONROM 0x30
+
+/** this is the master cbfs header - it need to be
+    located somewhere in the bootblock.  Where it
+    actually lives is up to coreboot. A pointer to
+    this header will live at 0xFFFFFFFc, so we can
+    easily find it. */
+
+#define CBFS_HEADER_MAGIC  0x4F524243
+#define CBFS_HEADPTR_ADDR 0xFFFFFFFc
+#define VERSION1 0x31313131
+
+struct cbfs_header {
+	u32 magic;
+        u32 version; 
+        u32 romsize;
+        u32 bootblocksize;
+	u32 align;
+	u32 offset;
+	u32 pad[2];
+} __attribute__((packed));
+
+/** This is a component header - every entry in the CBFS
+    will have this header.
+
+    This is how the component is arranged in the ROM:
+
+    --------------   <- 0
+    component header
+    --------------   <- sizeof(struct component)
+    component name
+    --------------   <- offset
+    data
+    ...
+    --------------   <- offset + len
+*/
+
+#define CBFS_FILE_MAGIC "LARCHIVE"
+
+struct cbfs_file {
+	char magic[8];
+	u32 len;
+	u32 type;
+	u32 checksum;
+	u32 offset;
+} __attribute__((packed));
+
+/*** Component sub-headers ***/
+
+/* Following are component sub-headers for the "standard"
+   component types */
+
+/** This is the sub-header for stage components.  Stages are
+    loaded by coreboot during the normal boot process */
+
+struct cbfs_stage {
+	u32 compression;  /** Compression type */
+	u64 entry;  /** entry point */
+	u64 load;   /** Where to load in memory */
+	u32 len;          /** length of data to load */
+	u32 memlen;	   /** total length of object in memory */
+} __attribute__((packed));
+
+/** this is the sub-header for payload components.  Payloads
+    are loaded by coreboot at the end of the boot process */
+
+struct cbfs_payload_segment {
+	u32 type;
+	u32 compression;
+	u32 offset;
+	u64 load_addr;
+	u32 len;
+	u32 mem_len;
+} __attribute__((packed));
+
+struct cbfs_payload {
+	struct cbfs_payload_segment segments;
+};
+
+#define PAYLOAD_SEGMENT_CODE   0x45444F43
+#define PAYLOAD_SEGMENT_DATA   0x41544144
+#define PAYLOAD_SEGMENT_BSS    0x20535342
+#define PAYLOAD_SEGMENT_PARAMS 0x41524150
+#define PAYLOAD_SEGMENT_ENTRY  0x52544E45
+
+struct cbfs_optionrom {
+	u32 compression;
+	u32 len;
+} __attribute__((packed));
+
+#define CBFS_NAME(_c) (((char *) (_c)) + sizeof(struct cbfs_file))
+#define CBFS_SUBHEADER(_p) ( (void *) ((((u8 *) (_p)) + ntohl((_p)->offset))) )
+
+void * cbfs_load_payload(struct cb_memory *cb_mem, const char *name);
+void * cbfs_load_stage(const char *name);
+int cbfs_execute_stage(const char *name);
+void * cbfs_get_file(const char *name);
+void *cbfs_load_optionrom(u16 vendor, u16 device, void * dest);
+int run_address(void *f);
+int cbfs_decompress(int algo, void *src, void *dst, int len);
+struct cbfs_stage *cbfs_find_file(const char *name, int type);
+int cbfs_check_magic(struct cbfs_file *file);
+struct cbfs_header *cbfs_master_header(void);
+struct cbfs_file *cbfs_find(const char *name);
+void cbfs_and_run_core(char* filename, unsigned int ebp);
+
+#endif
+
Index: include/read_bytes.h
===================================================================
--- include/read_bytes.h	(revision 0)
+++ include/read_bytes.h	(revision 7)
@@ -0,0 +1,11 @@
+#ifndef STREAM_READ_BYTES_H
+#define STREAM_READ_BYTES_H
+
+typedef long byte_offset_t;
+
+extern int stream_init(void);
+extern byte_offset_t stream_read(void *vdest, byte_offset_t count);
+extern byte_offset_t stream_skip(byte_offset_t count);
+extern void stream_fini(void);
+
+#endif /* STREAM_READ_BYTES_H */
Index: include/fallback_boot.h
===================================================================
--- include/fallback_boot.h	(revision 0)
+++ include/fallback_boot.h	(revision 7)
@@ -0,0 +1,18 @@
+#ifndef PART_FALLBACK_BOOT_H
+#define PART_FALLBACK_BOOT_H
+
+#ifndef ASSEMBLY
+
+#if CONFIG_HAVE_FALLBACK_BOOT == 1
+void set_boot_successful(void);
+#else
+#define set_boot_successful()
+#endif
+
+void boot_successful(void);
+
+#endif /* ASSEMBLY */
+
+#define RTC_BOOT_BYTE	48
+
+#endif /* PART_FALLBACK_BOOT_H */
Index: main/grub/builtins.c
===================================================================
--- main/grub/builtins.c	(revision 2)
+++ main/grub/builtins.c	(revision 7)
@@ -18,7 +18,10 @@
  * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
  */
 
+#define DEBUG_THIS 1
+
 #include <libpayload-config.h>
+#include <coreboot_tables.h>
 #include <libpayload.h>
 #include <config.h>
 #include <fs.h>
@@ -809,6 +812,69 @@
 };
 #endif
 
+#ifdef CONFIG_CBFS_BOOT
+#define BYTES_TO_MAP (1024*1024)
+
+static int payload_func(char *arg, int flags)
+{
+	unsigned char *table_area;
+	unsigned long addr, start;
+	struct cb_header *cb_table;
+	struct cb_record *rec, *end;
+
+	table_area = (unsigned char *) phys_to_virt(0x00000000);
+	cb_table = find_cb_table(table_area, 0x00000, 0x1000);
+	if (!cb_table)
+		cb_table = find_cb_table(table_area, 0xf0000, BYTES_TO_MAP);
+	if (cb_table) {
+		struct cb_forward *forward = (struct cb_forward *)
+			(((char *)cb_table) + cb_table->header_bytes);
+		if (forward->tag == CB_TAG_FORWARD) {
+			grub_printf("Found Coreboot High Tables at %p\r\n", forward);
+			start = forward->forward;
+			table_area = (unsigned char *) phys_to_virt(start);
+			cb_table = find_cb_table(table_area, 0x00000, 0x1000);
+		}
+	}
+
+	if (!cb_table) {
+		grub_printf("ERROR: No coreboot table found.\n");
+		return -1;
+	}
+	
+	rec = cb_first_record(cb_table);
+	end = cb_last_record(cb_table);
+	while (1) {
+		if (rec->tag == CB_TAG_MEMORY)
+		{
+			break;
+		}
+		else if (rec == end)
+		{
+			grub_printf("ERROR: No Coreboot Memory Map Found\n");
+			return -1;
+		}
+		else
+			rec = cb_next_record(rec);
+	}
+	
+	grub_printf("Loading CBFS Payload <%s>::\n\n", arg);
+	cbfs_load_payload((struct cb_memory *) rec, arg);
+	
+	return 0;
+}
+
+static struct builtin builtin_payload = {
+	"payload",
+	payload_func,
+	BUILTIN_CMDLINE | BUILTIN_HELP_LIST,
+	"payload FILE",
+	"Attempt to boot FILE as a CBFS (selfboot) payload"
+};
+
+
+#endif
+
 /* kernel */
 static int kernel_func(char *arg, int flags)
 {
@@ -1726,6 +1792,9 @@
 #endif
 	&builtin_password,
 	&builtin_pause,
+#ifdef CONFIG_CBFS_BOOT
+	&builtin_payload,
+#endif
 	&builtin_poweroff,
 #ifdef CONFIG_DEVELOPER_TOOLS
 	&builtin_probe,
Index: main/coreboot_table.c
===================================================================
--- main/coreboot_table.c	(revision 0)
+++ main/coreboot_table.c	(revision 7)
@@ -0,0 +1,121 @@
+#include <ipchecksum.h>
+#include <coreboot_tables.h>
+#include <coreboot_table.h>
+#include <debug.h>
+
+#define for_each_cbrec(head, rec) \
+	for(rec = (struct cb_record *)(((char *)head) + sizeof(*head)); \
+		(((char *)rec) < (((char *)head) + sizeof(*head) + head->table_bytes))  && \
+		(rec->size >= 1) && \
+		((((char *)rec) + rec->size) <= (((char *)head) + sizeof(*head) + head->table_bytes)); \
+		rec = (struct cb_record *)(((char *)rec) + rec->size))
+
+static unsigned long compute_checksum(void *addr, unsigned long length)
+{
+	uint8_t *ptr;
+	volatile union {
+		uint8_t byte[2];
+		uint16_t word;
+	} value;
+	unsigned long sum;
+	unsigned long i;
+
+	/* In the most straight forward way possible,
+	 * compute an ip style checksum.
+	 */
+	sum = 0;
+	ptr = addr;
+	for (i = 0; i < length; i++) {
+		unsigned long value2;
+		value2 = ptr[i];
+		if (i & 1) {
+			value2 <<= 8;
+		}
+		/* Add the new value */
+		sum += value2;
+		/* Wrap around the carry */
+		if (sum > 0xFFFF) {
+			sum = (sum + (sum >> 16)) & 0xFFFF;
+		}
+	}
+	value.byte[0] = sum & 0xff;
+	value.byte[1] = (sum >> 8) & 0xff;
+
+	return (~value.word) & 0xFFFF;
+}
+
+static int count_cb_records(struct cb_header *head)
+{
+	struct cb_record *rec;
+	int count;
+
+	count = 0;
+	for_each_cbrec(head, rec) {
+		count++;
+	}
+
+	return count;
+}
+
+struct cb_record *cb_first_record(struct cb_header *header)
+{
+	struct cb_record *rec;
+	rec = (void *)(((char *)header) + sizeof(*header));
+	return rec;
+}
+
+struct cb_record *cb_last_record(struct cb_header *header)
+{
+	struct cb_record *rec;
+	rec = (void *)(((char *)header) + sizeof(*header) + header->table_bytes);
+	return rec;
+}
+
+struct cb_record *cb_next_record(struct cb_record *rec)
+{
+	rec = (void *)(((char *)rec) + rec->size);	
+	return rec;
+}
+
+static struct cb_header *find_cb_table(void *base, unsigned long start,
+				       unsigned long end)
+{
+	unsigned long addr;
+
+	/* For now be stupid.... */
+	for (addr = start; addr < end; addr += 16) {
+		struct cb_header *head =
+		    (struct cb_header *)(((char *)base) + addr);
+		struct cb_record *recs =
+		    (struct cb_record *)(((char *)base) + addr + sizeof(*head));
+		if (memcmp(head->signature, "LBIO", 4) != 0)
+			continue;
+		debug("Found candidate at: %08lx-%08lx\n",
+			     addr, addr + head->table_bytes);
+		if (head->header_bytes != sizeof(*head)) {
+			debug("Header bytes of %d are incorrect.\n",
+				head->header_bytes);
+			continue;
+		}
+		if (count_cb_records(head) != head->table_entries) {
+			debug("Bad record count: %d.\n",
+				head->table_entries);
+			continue;
+		}
+		if (compute_checksum((uint8_t *) head, sizeof(*head)) != 0) {
+			debug("Bad header checksum.\n");
+			continue;
+		}
+		if (compute_checksum(recs, head->table_bytes)
+		    != head->table_checksum) {
+			debug("Bad table checksum: %04x.\n",
+				head->table_checksum);
+			continue;
+		}
+		debug("Found coreboot table at 0x%08lx.\n", addr);
+		return head;
+
+	};
+
+	return 0;
+}
Index: main/Makefile.inc
===================================================================
--- main/Makefile.inc	(revision 2)
+++ main/Makefile.inc	(revision 7)
@@ -19,4 +19,5 @@
 TARGETS-y += main/filo.o main/strtox.o
 TARGETS-y += main/elfload.o main/elfnote.o main/ipchecksum.o 
 TARGETS-$(CONFIG_SUPPORT_SOUND) += main/sound.o
+TARGETS-$(CONFIG_CBFS_BOOT) += main/selfboot.o 
 
Index: main/selfboot.c
===================================================================
--- main/selfboot.c	(revision 0)
+++ main/selfboot.c	(revision 7)
@@ -0,0 +1,318 @@
+/*
+ * This file is part of the coreboot project.
+ *
+ * Copyright (C) 2003 Eric W. Biederman <ebiederm@xmission.com>
+ * Copyright (C) 2009 Ron Minnich <rminnich@gmail.com>
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; version 2 of the License.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA, 02110-1301 USA
+ */
+
+#include <libpayload.h>
+#include <fallback_boot.h>
+#include <elf.h>
+#include <arch/elf.h>
+#include <elf_boot.h>
+#include <coreboot_tables.h>
+#include <ipchecksum.h>
+#include <read_bytes.h>
+#include <cbfs.h>
+#define DEBUG_THIS CONFIG_DEBUG_CBFSBOOT
+#include <debug.h>
+
+#ifndef CONFIG_BIG_ENDIAN
+#define ntohl(x) ( ((x&0xff)<<24) | ((x&0xff00)<<8) | \
+		((x&0xff0000) >> 8) | ((x&0xff000000) >> 24) )
+#else
+#define ntohl(x) (x)
+#endif
+
+/* Maximum physical address we can use for the coreboot bounce buffer.
+ */
+#ifndef MAX_ADDR
+#define MAX_ADDR -1UL
+#endif
+
+extern unsigned char _ram_seg;
+extern unsigned char _eram_seg;
+
+struct segment {
+	struct segment *next;
+	struct segment *prev;
+	struct segment *phdr_next;
+	struct segment *phdr_prev;
+	unsigned long s_dstaddr;
+	unsigned long s_srcaddr;
+	unsigned long s_memsz;
+	unsigned long s_filesz;
+	int compression;
+};
+
+struct verify_callback {
+	struct verify_callback *next;
+	int (*callback)(struct verify_callback *vcb, 
+		Elf_ehdr *ehdr, Elf_phdr *phdr, struct segment *head);
+	unsigned long desc_offset;
+	unsigned long desc_addr;
+};
+
+struct ip_checksum_vcb {
+	struct verify_callback data;
+	unsigned short ip_checksum;
+};
+
+void * cbfs_load_payload(struct cb_memory *cb_mem, const char *name)
+{
+	int selfboot(struct cb_memory *mem, struct cbfs_payload *payload);
+	struct cbfs_payload *payload;
+
+	payload = (struct cbfs_payload *)cbfs_find_file(name, CBFS_TYPE_PAYLOAD);
+	if (payload == NULL)
+		return (void *) -1;
+	debug("Got a payload\r\n");
+
+	selfboot(cb_mem, payload);
+	debug("SELFBOOT RETURNED!\r\n");
+
+	return (void *) -1;
+}
+
+static int valid_area(struct cb_memory *mem, unsigned long start, unsigned long len)
+{
+	/* Check through all of the memory segments and ensure
+	 * the segment that was passed in is completely contained
+	 * in RAM.
+	 */
+	int i;
+	unsigned long end = start + len;
+	unsigned long mem_entries = (mem->size - sizeof(*mem))/sizeof(mem->map[0]);
+
+	/* Walk through the table of valid memory ranges and see if I
+	 * have a match.
+	 */
+	for(i = 0; i < mem_entries; i++) {
+		uint64_t mstart, mend;
+		uint32_t mtype;
+		mtype = mem->map[i].type;
+		mstart = UNPACK_CB64(mem->map[i].start);
+		mend = mstart + UNPACK_CB64(mem->map[i].size);
+		if ((mtype == CB_MEM_RAM) && (start < mend) && (end > mstart)) {
+			break;
+		}
+		if ((mtype == CB_MEM_TABLE) && (start < mend) && (end > mstart)) {
+			debug("Payload is overwriting Coreboot tables.\r\n");
+			break;
+		}
+	}
+	if (i == mem_entries) {
+		debug("No matching ram area found for range:\r\n");
+		debug("  [0x%016lx, 0x%016lx)\r\n", start, end);
+		debug("Ram areas\r\n");
+		for(i = 0; i < mem_entries; i++) {
+			uint64_t mstart, mend;
+			uint32_t mtype;
+			mtype = mem->map[i].type;
+			mstart = UNPACK_CB64(mem->map[i].start);
+			mend = mstart + UNPACK_CB64(mem->map[i].size);
+			debug("  [0x%016lx, 0x%016lx) %s\r\n",
+				(unsigned long)mstart, 
+				(unsigned long)mend, 
+				(mtype == CB_MEM_RAM)?"RAM":"Reserved");
+			
+		}
+		return 0;
+	}
+	return 1;
+}
+
+static int build_self_segment_list(
+	struct segment *head, 
+	struct cb_memory *mem,
+	struct cbfs_payload *payload, u32 *entry)
+{
+	struct segment *new;
+	struct segment *ptr;
+	struct cbfs_payload_segment *segment, *first_segment;
+	memset(head, 0, sizeof(*head));
+	head->phdr_next = head->phdr_prev = head;
+	head->next = head->prev = head;
+	first_segment = segment = &payload->segments;
+
+	while(1) {
+		debug("Loading segment from rom address 0x%p\r\n", segment);
+		switch(segment->type) {
+		case PAYLOAD_SEGMENT_PARAMS:
+			debug("  parameter section (skipped)\r\n");
+			segment++;
+			continue;
+
+		case PAYLOAD_SEGMENT_CODE:
+		case PAYLOAD_SEGMENT_DATA:
+			debug("  %s (compression=%x)\r\n", 
+					segment->type == PAYLOAD_SEGMENT_CODE ?  "code" : "data",
+					ntohl(segment->compression));
+			new = malloc(sizeof(*new));
+			new->s_dstaddr = ntohl((u32) segment->load_addr);
+			new->s_memsz = ntohl(segment->mem_len);
+			new->compression = ntohl(segment->compression);
+
+			new->s_srcaddr = (u32) ((unsigned char *) first_segment) + ntohl(segment->offset);
+			new->s_filesz = ntohl(segment->len);
+			debug("  New segment dstaddr 0x%lx memsize 0x%lx srcaddr 0x%lx filesize 0x%lx\r\n",
+				new->s_dstaddr, new->s_memsz, new->s_srcaddr, new->s_filesz);
+			/* Clean up the values */
+			if (new->s_filesz > new->s_memsz)  {
+				new->s_filesz = new->s_memsz;
+			}
+			debug("  (cleaned up) New segment addr 0x%lx size 0x%lx offset 0x%lx filesize 0x%lx\r\n",
+				new->s_dstaddr, new->s_memsz, new->s_srcaddr, new->s_filesz);
+			break;
+
+		case PAYLOAD_SEGMENT_BSS:
+			debug("  BSS 0x%p (%d byte)\r\n", (void *) ntohl((u32) segment->load_addr),
+				 ntohl(segment->mem_len));
+			new = malloc(sizeof(*new));
+			new->s_filesz = 0;
+			new->s_dstaddr = ntohl((u32) segment->load_addr);
+			new->s_memsz = ntohl(segment->mem_len);
+			break;
+
+		case PAYLOAD_SEGMENT_ENTRY:
+			debug("  Entry Point 0x%p\r\n", (void *) ntohl((u32) segment->load_addr));
+			*entry =  ntohl((u32) segment->load_addr);
+			/* Per definition, a payload always has the entry point
+			 * as last segment. Thus, we use the occurence of the
+			 * entry point as break condition for the loop.
+			 * Can we actually just look at the number of section?
+			 */
+			return 1;
+
+		default:
+			/* We found something that we don't know about. Throw
+			 * hands into the sky and run away!
+			 */
+			debug("Bad segment type %x\r\n", segment->type);
+			return -1;
+		}
+
+		segment++;
+
+		for(ptr = head->next; ptr != head; ptr = ptr->next) {
+			if (new->s_srcaddr < ntohl((u32) segment->load_addr))
+				break;
+		}
+
+		/* Order by stream offset */
+		new->next = ptr;
+		new->prev = ptr->prev;
+		ptr->prev->next = new;
+		ptr->prev = new;
+
+		/* Order by original program header order */
+		new->phdr_next = head;
+		new->phdr_prev = head->phdr_prev;
+		head->phdr_prev->phdr_next  = new;
+		head->phdr_prev = new;
+	}
+
+	return 1;
+}
+
+static int load_self_segments(
+	struct segment *head,
+	struct cb_memory *mem,
+	struct cbfs_payload *payload)
+{
+	struct segment *ptr;
+	
+	for(ptr = head->next; ptr != head; ptr = ptr->next) {
+		/* Verify the memory addresses in the segment are valid */
+		if (!valid_area(mem, ptr->s_dstaddr, ptr->s_memsz))
+			return 0;
+	}
+	for(ptr = head->next; ptr != head; ptr = ptr->next) {
+		unsigned char *dest, *src;
+		debug("Loading Segment: addr: 0x%016lx memsz: 0x%016lx filesz: 0x%016lx\r\n",
+			ptr->s_dstaddr, ptr->s_memsz, ptr->s_filesz);
+		
+		/* Compute the boundaries of the segment */
+		dest = (unsigned char *)phys_to_virt((ptr->s_dstaddr));
+		src = (unsigned char *)(ptr->s_srcaddr);
+		
+		/* Copy data from the initial buffer */
+		if (ptr->s_filesz) {
+			unsigned char *middle, *end;
+			size_t len;
+			len = ptr->s_filesz;
+			switch(ptr->compression) {
+				case CBFS_COMPRESS_LZMA: {
+					debug("using LZMA\r\n");
+					unsigned long ulzma(unsigned char *src, unsigned char *dst);		
+					len = ulzma(src, dest);
+					break;
+				}
+				case CBFS_COMPRESS_NONE: {
+					debug("it's not compressed!\r\n");
+					memcpy(dest, src, len);
+					break;
+				}
+				default:
+					debug( "CBFS:  Unknown compression type %d\r\n", ptr->compression);
+					return -1;
+			}
+			end = dest + ptr->s_memsz;
+			middle = dest + len;
+			debug("[ 0x%016lx, %016lx, 0x%016lx) <- %016lx\r\n",
+				(unsigned long)dest,
+				(unsigned long)middle,
+				(unsigned long)end,
+				(unsigned long)src);
+
+			/* Zero the extra bytes between middle & end */
+			if (middle < end) {
+				debug("Clearing Segment: addr: 0x%016lx memsz: 0x%016lx\r\n",
+					(unsigned long)middle, (unsigned long)(end - middle));
+			
+				/* Zero the extra bytes */
+				memset(middle, 0, end - middle);
+			}
+		}
+	}
+	return 1;
+}
+
+int selfboot(struct cb_memory *mem, struct cbfs_payload *payload)
+{
+	u32 entry=0;
+	struct segment head;
+
+	/* Preprocess the self segments */
+	if (!build_self_segment_list(&head, mem, payload, &entry))
+		goto out;
+
+	/* Load the segments */
+	if (!load_self_segments(&head, mem, payload))
+		goto out;
+
+	debug("Loaded segments\r\n");
+
+	debug("Jumping to boot code at %x\r\n", entry);
+
+	/* Jump to kernel */
+	jmp_to_elf_entry((void*)entry);
+	return 1;
+
+ out:
+	return 0;
+}
+
Index: fs/vfs.c
===================================================================
--- fs/vfs.c	(revision 2)
+++ fs/vfs.c	(revision 7)
@@ -47,6 +47,9 @@
 };
 
 struct fsys_entry fsys_table[] = {
+# ifdef CONFIG_FSYS_CBFS
+	{"CBFS ROM Image", cbfs_mount, cbfs_read, cbfs_dir, 0, 0},
+# endif
 # ifdef CONFIG_FSYS_FAT
 	{"FAT filesystem", fat_mount, fat_read, fat_dir, 0, 0},
 # endif
Index: fs/lzma.c
===================================================================
--- fs/lzma.c	(revision 0)
+++ fs/lzma.c	(revision 7)
@@ -0,0 +1,47 @@
+/* 
+
+Coreboot interface to memory-saving variant of LZMA decoder
+
+(C)opyright 2006 Carl-Daniel Hailfinger
+Released under the GNU GPL v2 or later
+
+Parts of this file are based on C/7zip/Compress/LZMA_C/LzmaTest.c from the LZMA
+SDK 4.42, which is written and distributed to public domain by Igor Pavlov.
+
+*/
+
+#include "lzmadecode.c"
+#define printk_warning printf
+
+
+unsigned long ulzma(unsigned char * src, unsigned char * dst)
+{
+	unsigned char properties[LZMA_PROPERTIES_SIZE];
+	UInt32 outSize;
+	SizeT inProcessed;
+	SizeT outProcessed;
+	int res;
+	CLzmaDecoderState state;
+	SizeT mallocneeds;
+	unsigned char scratchpad[15980];
+
+	memcpy(properties, src, LZMA_PROPERTIES_SIZE);
+	outSize = *(UInt32 *)(src + LZMA_PROPERTIES_SIZE);
+	if (LzmaDecodeProperties(&state.Properties, properties, LZMA_PROPERTIES_SIZE) != LZMA_RESULT_OK) {
+		printk_warning("lzma: Incorrect stream properties.\n");
+		return 0;
+	}
+	mallocneeds = (LzmaGetNumProbs(&state.Properties) * sizeof(CProb));
+	if (mallocneeds > 15980) {
+		printk_warning("lzma: Decoder scratchpad too small!\n");
+		return 0;
+	}
+	state.Probs = (CProb *)scratchpad;
+	res = LzmaDecode(&state, src + LZMA_PROPERTIES_SIZE + 8, (SizeT)0xffffffff, &inProcessed,
+		dst, outSize, &outProcessed);
+	if (res != 0) {
+		printk_warning("lzma: Decoding error = %d\n", res);
+		return 0;
+	}
+	return outSize;
+}
Index: fs/Makefile.inc
===================================================================
--- fs/Makefile.inc	(revision 2)
+++ fs/Makefile.inc	(revision 7)
@@ -30,4 +30,6 @@
 TARGETS-$(CONFIG_FSYS_SQUASHFS) += fs/fsys_squashfs.o
 TARGETS-$(CONFIG_FSYS_SQUASHFS) += fs/squashfs_zlib.o
 TARGETS-$(CONFIG_ARTEC_BOOT) += fs/fsys_aboot.o
-
+TARGETS-$(CONFIG_FSYS_CBFS) += fs/fsys_cbfs.o
+TARGETS-$(CONFIG_FSYS_CBFS) += fs/cbfs.o
+TARGETS-$(CONFIG_FSYS_CBFS) += fs/lzma.o
Index: fs/lzmadecode.c
===================================================================
--- fs/lzmadecode.c	(revision 0)
+++ fs/lzmadecode.c	(revision 7)
@@ -0,0 +1,398 @@
+/*
+  LzmaDecode.c
+  LZMA Decoder (optimized for Speed version)
+  
+  LZMA SDK 4.40 Copyright (c) 1999-2006 Igor Pavlov (2006-05-01)
+  http://www.7-zip.org/
+
+  LZMA SDK is licensed under two licenses:
+  1) GNU Lesser General Public License (GNU LGPL)
+  2) Common Public License (CPL)
+  It means that you can select one of these two licenses and 
+  follow rules of that license.
+
+  SPECIAL EXCEPTION:
+  Igor Pavlov, as the author of this Code, expressly permits you to 
+  statically or dynamically link your Code (or bind by name) to the 
+  interfaces of this file without subjecting your linked Code to the 
+  terms of the CPL or GNU LGPL. Any modifications or additions 
+  to this file, however, are subject to the LGPL or CPL terms.
+*/
+
+#include "lzmadecode.h"
+
+#define kNumTopBits 24
+#define kTopValue ((UInt32)1 << kNumTopBits)
+
+#define kNumBitModelTotalBits 11
+#define kBitModelTotal (1 << kNumBitModelTotalBits)
+#define kNumMoveBits 5
+
+#define RC_READ_BYTE (*Buffer++)
+
+#define RC_INIT2 Code = 0; Range = 0xFFFFFFFF; \
+  { int i; for(i = 0; i < 5; i++) { RC_TEST; Code = (Code << 8) | RC_READ_BYTE; }}
+
+
+#define RC_TEST { if (Buffer == BufferLim) return LZMA_RESULT_DATA_ERROR; }
+
+#define RC_INIT(buffer, bufferSize) Buffer = buffer; BufferLim = buffer + bufferSize; RC_INIT2
+ 
+
+#define RC_NORMALIZE if (Range < kTopValue) { RC_TEST; Range <<= 8; Code = (Code << 8) | RC_READ_BYTE; }
+
+#define IfBit0(p) RC_NORMALIZE; bound = (Range >> kNumBitModelTotalBits) * *(p); if (Code < bound)
+#define UpdateBit0(p) Range = bound; *(p) += (kBitModelTotal - *(p)) >> kNumMoveBits;
+#define UpdateBit1(p) Range -= bound; Code -= bound; *(p) -= (*(p)) >> kNumMoveBits;
+
+#define RC_GET_BIT2(p, mi, A0, A1) IfBit0(p) \
+  { UpdateBit0(p); mi <<= 1; A0; } else \
+  { UpdateBit1(p); mi = (mi + mi) + 1; A1; } 
+  
+#define RC_GET_BIT(p, mi) RC_GET_BIT2(p, mi, ; , ;)               
+
+#define RangeDecoderBitTreeDecode(probs, numLevels, res) \
+  { int i = numLevels; res = 1; \
+  do { CProb *cp = probs + res; RC_GET_BIT(cp, res) } while(--i != 0); \
+  res -= (1 << numLevels); }
+
+
+#define kNumPosBitsMax 4
+#define kNumPosStatesMax (1 << kNumPosBitsMax)
+
+#define kLenNumLowBits 3
+#define kLenNumLowSymbols (1 << kLenNumLowBits)
+#define kLenNumMidBits 3
+#define kLenNumMidSymbols (1 << kLenNumMidBits)
+#define kLenNumHighBits 8
+#define kLenNumHighSymbols (1 << kLenNumHighBits)
+
+#define LenChoice 0
+#define LenChoice2 (LenChoice + 1)
+#define LenLow (LenChoice2 + 1)
+#define LenMid (LenLow + (kNumPosStatesMax << kLenNumLowBits))
+#define LenHigh (LenMid + (kNumPosStatesMax << kLenNumMidBits))
+#define kNumLenProbs (LenHigh + kLenNumHighSymbols) 
+
+
+#define kNumStates 12
+#define kNumLitStates 7
+
+#define kStartPosModelIndex 4
+#define kEndPosModelIndex 14
+#define kNumFullDistances (1 << (kEndPosModelIndex >> 1))
+
+#define kNumPosSlotBits 6
+#define kNumLenToPosStates 4
+
+#define kNumAlignBits 4
+#define kAlignTableSize (1 << kNumAlignBits)
+
+#define kMatchMinLen 2
+
+#define IsMatch 0
+#define IsRep (IsMatch + (kNumStates << kNumPosBitsMax))
+#define IsRepG0 (IsRep + kNumStates)
+#define IsRepG1 (IsRepG0 + kNumStates)
+#define IsRepG2 (IsRepG1 + kNumStates)
+#define IsRep0Long (IsRepG2 + kNumStates)
+#define PosSlot (IsRep0Long + (kNumStates << kNumPosBitsMax))
+#define SpecPos (PosSlot + (kNumLenToPosStates << kNumPosSlotBits))
+#define Align (SpecPos + kNumFullDistances - kEndPosModelIndex)
+#define LenCoder (Align + kAlignTableSize)
+#define RepLenCoder (LenCoder + kNumLenProbs)
+#define Literal (RepLenCoder + kNumLenProbs)
+
+#if Literal != LZMA_BASE_SIZE
+StopCompilingDueBUG
+#endif
+
+int LzmaDecodeProperties(CLzmaProperties *propsRes, const unsigned char *propsData, int size)
+{
+  unsigned char prop0;
+  if (size < LZMA_PROPERTIES_SIZE)
+    return LZMA_RESULT_DATA_ERROR;
+  prop0 = propsData[0];
+  if (prop0 >= (9 * 5 * 5))
+    return LZMA_RESULT_DATA_ERROR;
+  {
+    for (propsRes->pb = 0; prop0 >= (9 * 5); propsRes->pb++, prop0 -= (9 * 5));
+    for (propsRes->lp = 0; prop0 >= 9; propsRes->lp++, prop0 -= 9);
+    propsRes->lc = prop0;
+    /*
+    unsigned char remainder = (unsigned char)(prop0 / 9);
+    propsRes->lc = prop0 % 9;
+    propsRes->pb = remainder / 5;
+    propsRes->lp = remainder % 5;
+    */
+  }
+
+  return LZMA_RESULT_OK;
+}
+
+#define kLzmaStreamWasFinishedId (-1)
+
+int LzmaDecode(CLzmaDecoderState *vs,
+    const unsigned char *inStream, SizeT inSize, SizeT *inSizeProcessed,
+    unsigned char *outStream, SizeT outSize, SizeT *outSizeProcessed)
+{
+  CProb *p = vs->Probs;
+  SizeT nowPos = 0;
+  Byte previousByte = 0;
+  UInt32 posStateMask = (1 << (vs->Properties.pb)) - 1;
+  UInt32 literalPosMask = (1 << (vs->Properties.lp)) - 1;
+  int lc = vs->Properties.lc;
+
+
+  int state = 0;
+  UInt32 rep0 = 1, rep1 = 1, rep2 = 1, rep3 = 1;
+  int len = 0;
+  const Byte *Buffer;
+  const Byte *BufferLim;
+  UInt32 Range;
+  UInt32 Code;
+
+  *inSizeProcessed = 0;
+  *outSizeProcessed = 0;
+
+  {
+    UInt32 i;
+    UInt32 numProbs = Literal + ((UInt32)LZMA_LIT_SIZE << (lc + vs->Properties.lp));
+    for (i = 0; i < numProbs; i++)
+      p[i] = kBitModelTotal >> 1;
+  }
+  
+  RC_INIT(inStream, inSize);
+
+
+  while(nowPos < outSize)
+  {
+    CProb *prob;
+    UInt32 bound;
+    int posState = (int)(
+        (nowPos 
+        )
+        & posStateMask);
+
+    prob = p + IsMatch + (state << kNumPosBitsMax) + posState;
+    IfBit0(prob)
+    {
+      int symbol = 1;
+      UpdateBit0(prob)
+      prob = p + Literal + (LZMA_LIT_SIZE * 
+        (((
+        (nowPos 
+        )
+        & literalPosMask) << lc) + (previousByte >> (8 - lc))));
+
+      if (state >= kNumLitStates)
+      {
+        int matchByte;
+        matchByte = outStream[nowPos - rep0];
+        do
+        {
+          int bit;
+          CProb *probLit;
+          matchByte <<= 1;
+          bit = (matchByte & 0x100);
+          probLit = prob + 0x100 + bit + symbol;
+          RC_GET_BIT2(probLit, symbol, if (bit != 0) break, if (bit == 0) break)
+        }
+        while (symbol < 0x100);
+      }
+      while (symbol < 0x100)
+      {
+        CProb *probLit = prob + symbol;
+        RC_GET_BIT(probLit, symbol)
+      }
+      previousByte = (Byte)symbol;
+
+      outStream[nowPos++] = previousByte;
+      if (state < 4) state = 0;
+      else if (state < 10) state -= 3;
+      else state -= 6;
+    }
+    else             
+    {
+      UpdateBit1(prob);
+      prob = p + IsRep + state;
+      IfBit0(prob)
+      {
+        UpdateBit0(prob);
+        rep3 = rep2;
+        rep2 = rep1;
+        rep1 = rep0;
+        state = state < kNumLitStates ? 0 : 3;
+        prob = p + LenCoder;
+      }
+      else
+      {
+        UpdateBit1(prob);
+        prob = p + IsRepG0 + state;
+        IfBit0(prob)
+        {
+          UpdateBit0(prob);
+          prob = p + IsRep0Long + (state << kNumPosBitsMax) + posState;
+          IfBit0(prob)
+          {
+            UpdateBit0(prob);
+            
+            if (nowPos == 0)
+              return LZMA_RESULT_DATA_ERROR;
+            
+            state = state < kNumLitStates ? 9 : 11;
+            previousByte = outStream[nowPos - rep0];
+            outStream[nowPos++] = previousByte;
+
+            continue;
+          }
+          else
+          {
+            UpdateBit1(prob);
+          }
+        }
+        else
+        {
+          UInt32 distance;
+          UpdateBit1(prob);
+          prob = p + IsRepG1 + state;
+          IfBit0(prob)
+          {
+            UpdateBit0(prob);
+            distance = rep1;
+          }
+          else 
+          {
+            UpdateBit1(prob);
+            prob = p + IsRepG2 + state;
+            IfBit0(prob)
+            {
+              UpdateBit0(prob);
+              distance = rep2;
+            }
+            else
+            {
+              UpdateBit1(prob);
+              distance = rep3;
+              rep3 = rep2;
+            }
+            rep2 = rep1;
+          }
+          rep1 = rep0;
+          rep0 = distance;
+        }
+        state = state < kNumLitStates ? 8 : 11;
+        prob = p + RepLenCoder;
+      }
+      {
+        int numBits, offset;
+        CProb *probLen = prob + LenChoice;
+        IfBit0(probLen)
+        {
+          UpdateBit0(probLen);
+          probLen = prob + LenLow + (posState << kLenNumLowBits);
+          offset = 0;
+          numBits = kLenNumLowBits;
+        }
+        else
+        {
+          UpdateBit1(probLen);
+          probLen = prob + LenChoice2;
+          IfBit0(probLen)
+          {
+            UpdateBit0(probLen);
+            probLen = prob + LenMid + (posState << kLenNumMidBits);
+            offset = kLenNumLowSymbols;
+            numBits = kLenNumMidBits;
+          }
+          else
+          {
+            UpdateBit1(probLen);
+            probLen = prob + LenHigh;
+            offset = kLenNumLowSymbols + kLenNumMidSymbols;
+            numBits = kLenNumHighBits;
+          }
+        }
+        RangeDecoderBitTreeDecode(probLen, numBits, len);
+        len += offset;
+      }
+
+      if (state < 4)
+      {
+        int posSlot;
+        state += kNumLitStates;
+        prob = p + PosSlot +
+            ((len < kNumLenToPosStates ? len : kNumLenToPosStates - 1) << 
+            kNumPosSlotBits);
+        RangeDecoderBitTreeDecode(prob, kNumPosSlotBits, posSlot);
+        if (posSlot >= kStartPosModelIndex)
+        {
+          int numDirectBits = ((posSlot >> 1) - 1);
+          rep0 = (2 | ((UInt32)posSlot & 1));
+          if (posSlot < kEndPosModelIndex)
+          {
+            rep0 <<= numDirectBits;
+            prob = p + SpecPos + rep0 - posSlot - 1;
+          }
+          else
+          {
+            numDirectBits -= kNumAlignBits;
+            do
+            {
+              RC_NORMALIZE
+              Range >>= 1;
+              rep0 <<= 1;
+              if (Code >= Range)
+              {
+                Code -= Range;
+                rep0 |= 1;
+              }
+            }
+            while (--numDirectBits != 0);
+            prob = p + Align;
+            rep0 <<= kNumAlignBits;
+            numDirectBits = kNumAlignBits;
+          }
+          {
+            int i = 1;
+            int mi = 1;
+            do
+            {
+              CProb *prob3 = prob + mi;
+              RC_GET_BIT2(prob3, mi, ; , rep0 |= i);
+              i <<= 1;
+            }
+            while(--numDirectBits != 0);
+          }
+        }
+        else
+          rep0 = posSlot;
+        if (++rep0 == (UInt32)(0))
+        {
+          /* it's for stream version */
+          len = kLzmaStreamWasFinishedId;
+          break;
+        }
+      }
+
+      len += kMatchMinLen;
+      if (rep0 > nowPos)
+        return LZMA_RESULT_DATA_ERROR;
+
+
+      do
+      {
+        previousByte = outStream[nowPos - rep0];
+        len--;
+        outStream[nowPos++] = previousByte;
+      }
+      while(len != 0 && nowPos < outSize);
+    }
+  }
+  RC_NORMALIZE;
+
+
+  *inSizeProcessed = (SizeT)(Buffer - inStream);
+  *outSizeProcessed = nowPos;
+  return LZMA_RESULT_OK;
+}
Index: fs/lzmadecode.h
===================================================================
--- fs/lzmadecode.h	(revision 0)
+++ fs/lzmadecode.h	(revision 7)
@@ -0,0 +1,67 @@
+/* 
+  LzmaDecode.h
+  LZMA Decoder interface
+
+  LZMA SDK 4.40 Copyright (c) 1999-2006 Igor Pavlov (2006-05-01)
+  http://www.7-zip.org/
+
+  LZMA SDK is licensed under two licenses:
+  1) GNU Lesser General Public License (GNU LGPL)
+  2) Common Public License (CPL)
+  It means that you can select one of these two licenses and 
+  follow rules of that license.
+
+  SPECIAL EXCEPTION:
+  Igor Pavlov, as the author of this code, expressly permits you to 
+  statically or dynamically link your code (or bind by name) to the 
+  interfaces of this file without subjecting your linked code to the 
+  terms of the CPL or GNU LGPL. Any modifications or additions 
+  to this file, however, are subject to the LGPL or CPL terms.
+*/
+
+#ifndef __LZMADECODE_H
+#define __LZMADECODE_H
+
+typedef unsigned char Byte;
+typedef unsigned short UInt16;
+typedef unsigned int UInt32;
+typedef UInt32 SizeT;
+
+#define CProb UInt16
+
+#define LZMA_RESULT_OK 0
+#define LZMA_RESULT_DATA_ERROR 1
+
+
+#define LZMA_BASE_SIZE 1846
+#define LZMA_LIT_SIZE 768
+
+#define LZMA_PROPERTIES_SIZE 5
+
+typedef struct _CLzmaProperties
+{
+  int lc;
+  int lp;
+  int pb;
+}CLzmaProperties;
+
+int LzmaDecodeProperties(CLzmaProperties *propsRes, const unsigned char *propsData, int size);
+
+#define LzmaGetNumProbs(Properties) (LZMA_BASE_SIZE + (LZMA_LIT_SIZE << ((Properties)->lc + (Properties)->lp)))
+
+#define kLzmaNeedInitId (-2)
+
+typedef struct _CLzmaDecoderState
+{
+  CLzmaProperties Properties;
+  CProb *Probs;
+
+
+} CLzmaDecoderState;
+
+
+int LzmaDecode(CLzmaDecoderState *vs,
+    const unsigned char *inStream, SizeT inSize, SizeT *inSizeProcessed,
+    unsigned char *outStream, SizeT outSize, SizeT *outSizeProcessed);
+
+#endif
Index: fs/cbfs.c
===================================================================
--- fs/cbfs.c	(revision 0)
+++ fs/cbfs.c	(revision 7)
@@ -0,0 +1,236 @@
+/*
+ * This file is part of the coreboot project.
+ *
+ * Copyright (C) 2008, Jordan Crouse <jordan@cosmicpenguin.net>
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; version 2 of the License.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA, 02110-1301 USA
+ */
+
+#include <arch/types.h>
+#include <arch/virtual.h>
+#include <cbfs.h>
+
+#define ALIGN(val,by) ((val+by-1)&~(by-1))
+#define cbfs_printf(...) /* printf(__VA_ARGS__) */
+#define printk_spew cbfs_printf
+#define printk_debug cbfs_printf
+#define printk_info cbfs_printf
+#define printk_err cbfs_printf
+
+#ifndef CONFIG_BIG_ENDIAN
+#define ntohl(x) ( ((x&0xff)<<24) | ((x&0xff00)<<8) | \
+		((x&0xff0000) >> 8) | ((x&0xff000000) >> 24) )
+#else
+#define ntohl(x) (x)
+#endif
+
+int cbfs_decompress(int algo, void *src, void *dst, int len)
+{
+	switch(algo) {
+	case CBFS_COMPRESS_NONE:
+		memcpy(dst, src, len);
+		return 0;
+
+	case CBFS_COMPRESS_LZMA: {
+		unsigned long ulzma(unsigned char *src, unsigned char *dst);
+		ulzma(src, dst);
+	}
+		return 0;
+
+	default:
+		printk_info( "CBFS:  Unknown compression type %d\n",
+		       algo);
+		return -1;
+	}
+}
+
+int cbfs_check_magic(struct cbfs_file *file)
+{
+	return !strcmp(file->magic, CBFS_FILE_MAGIC) ? 1 : 0;
+}
+
+struct cbfs_header *cbfs_master_header(void)
+{
+	struct cbfs_header *header;
+
+	void *ptr = phys_to_virt((void *)*((unsigned long *) phys_to_virt(CBFS_HEADPTR_ADDR)));
+	printk_spew("Check CBFS header at %p\n", ptr);
+	header = (struct cbfs_header *) ptr;
+
+	printk_spew("magic is %08x\n", ntohl(header->magic));
+	if (ntohl(header->magic) != CBFS_HEADER_MAGIC) {
+		printk_err("ERROR: No valid CBFS header found!\n");
+		return NULL;
+	}
+
+	printk_spew("Found CBFS header at %p\n", ptr);
+	return header;
+}
+
+struct cbfs_file *cbfs_find(const char *name)
+{
+	struct cbfs_header *header = cbfs_master_header();
+	unsigned long offset;
+
+	if (header == NULL)
+		return NULL;
+	offset = (unsigned long) phys_to_virt(0 - ntohl(header->romsize) + ntohl(header->offset));
+
+	int align= ntohl(header->align);
+
+	while(1) {
+		struct cbfs_file *file = (struct cbfs_file *) offset;
+		if (!cbfs_check_magic(file)) return NULL;
+		printk_debug("Check %s\n", CBFS_NAME(file));
+		if (!strcmp(CBFS_NAME(file), name))
+			return file;
+
+		int flen = ntohl(file->len);
+		int foffset = ntohl(file->offset);
+		printk_spew("CBFS: follow chain: %p + %x + %x + align -> ", (void *)offset, foffset, flen);
+
+		unsigned long oldoffset = offset;
+		offset = ALIGN(offset + foffset + flen, align);
+		printk_spew("%p\n", (void *)offset);
+		if (virt_to_phys(offset) <= virt_to_phys(oldoffset)) return NULL;
+
+		if (offset < (unsigned long) phys_to_virt(0xFFFFFFFF - ntohl(header->romsize)))
+			return NULL;
+	}
+}
+
+struct cbfs_stage *cbfs_find_file(const char *name, int type)
+{
+	struct cbfs_file *file = cbfs_find(name);
+
+	if (file == NULL) {
+		printk_info( "CBFS:  Could not find file %s\n",
+		       name);
+		return NULL;
+	}
+
+	if (ntohl(file->type) != type) {
+		printk_info( "CBFS:  File %s is of type %x instead of"
+		       "type %x\n", name, file->type, type);
+
+		return NULL;
+	}
+
+	return (void *) CBFS_SUBHEADER(file);
+}
+
+void *cbfs_load_optionrom(u16 vendor, u16 device, void * dest)
+{
+	char name[17];
+	struct cbfs_optionrom *orom;
+	u8 *src;
+
+	sprintf(name,"pci%04x,%04x.rom", vendor, device);
+
+	orom = (struct cbfs_optionrom *)
+		cbfs_find_file(name, CBFS_TYPE_OPTIONROM);
+
+	if (orom == NULL)
+		return NULL;
+
+	/* They might have specified a dest address. If so, we can decompress. 
+	 * If not, there's not much hope of decompressing or relocating the rom.
+	 * in the common case, the expansion rom is uncompressed, we
+	 * pass 0 in for the dest, and all we have to do is find the rom and 
+	 * return a pointer to it. 
+ 	 */
+
+	/* BUG: the cbfstool is (not yet) including a cbfs_optionrom header */
+	src = ((unsigned char *) orom); // + sizeof(struct cbfs_optionrom);
+
+	if (! dest)
+		return src;
+
+	if (cbfs_decompress(ntohl(orom->compression),
+			     src,
+			     dest,
+			     ntohl(orom->len)))
+		return NULL;
+
+	return dest;
+}
+
+void * cbfs_load_stage(const char *name)
+{
+	struct cbfs_stage *stage = (struct cbfs_stage *)
+		cbfs_find_file(name, CBFS_TYPE_STAGE);
+	/* this is a mess. There is no ntohll. */
+	/* for now, assume compatible byte order until we solve this. */
+	u32 entry;
+
+	if (stage == NULL)
+		return (void *) -1;
+
+	printk_info("Stage: load %s @ %d/%d bytes, enter @ %llx\n", 
+			name,
+			(u32) stage->load, stage->memlen, 
+			stage->entry);
+	memset((void *) (u32) stage->load, 0, stage->memlen);
+
+	if (cbfs_decompress(stage->compression,
+			     ((unsigned char *) stage) +
+			     sizeof(struct cbfs_stage),
+			     (void *) (u32) stage->load,
+			     stage->len))
+		return (void *) -1;
+	printk_info("Stage: done loading.\n");
+
+	entry = stage->entry;
+//	return (void *) ntohl((u32) stage->entry);
+	return (void *) entry;
+}
+
+void * cbfs_get_file(const char *name)
+{
+	return cbfs_find(name);
+}
+
+int cbfs_execute_stage(const char *name)
+{
+	struct cbfs_stage *stage = (struct cbfs_stage *)
+		cbfs_find_file(name, CBFS_TYPE_STAGE);
+
+	if (stage == NULL)
+		return 1;
+
+	if (ntohl(stage->compression) != CBFS_COMPRESS_NONE) {
+		printk_info( "CBFS:  Unable to run %s:  Compressed file"
+		       "Not supported for in-place execution\n", name);
+		return 1;
+	}
+
+	/* FIXME: This isn't right */
+	printk_info( "CBFS: run @ %p\n", (void *) ntohl((u32) stage->entry));
+	return run_address((void *) ntohl((u32) stage->entry));
+}
+
+/**
+ * run_address is passed the address of a function taking no parameters and
+ * jumps to it, returning the result. 
+ * @param f the address to call as a function. 
+ * returns value returned by the function. 
+ */
+
+int run_address(void *f)
+{
+	int (*v) (void);
+	v = f;
+	return v();
+}
+
Index: fs/fsys_cbfs.c
===================================================================
--- fs/fsys_cbfs.c	(revision 0)
+++ fs/fsys_cbfs.c	(revision 7)
@@ -0,0 +1,66 @@
+/*
+ *  GRUB  --  GRand Unified Bootloader
+ *  Copyright (C) 2000,2001,2005   Free Software Foundation, Inc.
+ *  Copyright (C) 2009   coresystems GmbH
+ *
+ *  This program is free software; you can redistribute it and/or modify
+ *  it under the terms of the GNU General Public License, version 2,
+ *  as published by the Free Software Foundation.
+ *
+ *  This program is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *  GNU General Public License for more details.
+ *
+ *  You should have received a copy of the GNU General Public License
+ *  along with this program; if not, write to the Free Software
+ *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ */
+
+#include "filesys.h"
+#include <cbfs.h>
+
+struct cbfs_file *file;
+void* fileptr;
+
+int
+cbfs_mount (void)
+{
+  /* Check if it's a 4GB device (ie. RAM) */
+  if (part_length != (1<<(32-9)))
+    return 0;
+
+  /* CBFS? */
+  if (!cbfs_master_header())
+    return 0;
+
+  return 1;
+}
+
+int
+cbfs_read (char *buf, int len)
+{
+  if (filepos + len > filemax) {
+    len = filemax - filepos;
+  }
+  memcpy(buf, fileptr+filepos, len);
+  filepos+=len;
+  return len;
+}
+
+int
+cbfs_dir (char *dirname)
+{
+  if (dirname[0]=='/') dirname++;
+  file = cbfs_find(dirname);
+  if (!file) {
+    errnum = ERR_FILE_NOT_FOUND;
+    return 0;
+  }
+
+  filepos = 0;
+  filemax = ntohl(file->len);
+  fileptr = (void*)file+ntohl(file->offset);
+  return 1;
+}
+
Index: fs/filesys.h
===================================================================
--- fs/filesys.h	(revision 2)
+++ fs/filesys.h	(revision 7)
@@ -162,6 +162,12 @@
 int fat_dir (char *dirname);
 #endif
 
+#ifdef CONFIG_FSYS_CBFS
+int cbfs_mount (void);
+int cbfs_read (char *buf, int len);
+int cbfs_dir (char *dirname);
+#endif
+
 #ifdef CONFIG_FSYS_EXT2FS
 int ext2fs_mount (void);
 int ext2fs_read (char *buf, int len);
Index: i386/include/arch/multiboot.h
===================================================================
--- i386/include/arch/multiboot.h	(revision 0)
+++ i386/include/arch/multiboot.h	(revision 7)
@@ -0,0 +1,174 @@
+/* multiboot.h - multiboot header file. */
+/*
+ *  Copyright (C) 2003  Free Software Foundation, Inc.
+ *  Copyright (C) 2008  Robert Millan
+ *
+ *  This program is free software; you can redistribute it and/or modify
+ *  it under the terms of the GNU General Public License as published by
+ *  the Free Software Foundation; either version 2 of the License, or
+ *  (at your option) any later version.
+ *
+ *  This program is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *  GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, see <http://www.gnu.org/licenses/>.
+ */
+
+#ifndef MULTIBOOT_H
+#define MULTIBOOT_H
+
+/* How many bytes from the start of the file we search for the header.  */
+#define MB_SEARCH                 8192
+
+/* The magic field should contain this.  */
+#define MB_MAGIC                  0x1BADB002
+
+/* This should be in %eax.  */
+#define MB_MAGIC2                 0x2BADB002
+
+/* The bits in the required part of flags field we don't support.  */
+#define MB_UNSUPPORTED            0x0000fffc
+
+/* Alignment of multiboot modules.  */
+#define MB_MOD_ALIGN              0x00001000
+
+/*
+ * Flags set in the 'flags' member of the multiboot header.
+ */
+
+/* Align all boot modules on i386 page (4KB) boundaries.  */
+#define MB_PAGE_ALIGN		0x00000001
+
+/* Must pass memory information to OS.  */
+#define MB_MEMORY_INFO		0x00000002
+
+/* Must pass video information to OS.  */
+#define MB_VIDEO_MODE		0x00000004
+
+/* This flag indicates the use of the address fields in the header.  */
+#define MB_AOUT_KLUDGE		0x00010000
+
+/*
+ *  Flags to be set in the 'flags' member of the multiboot info structure.
+ */
+
+/* is there basic lower/upper memory information? */
+#define MB_INFO_MEMORY		0x00000001
+/* is there a boot device set? */
+#define MB_INFO_BOOTDEV		0x00000002
+/* is the command-line defined? */
+#define MB_INFO_CMDLINE		0x00000004
+/* are there modules to do something with? */
+#define MB_INFO_MODS		0x00000008
+
+/* These next two are mutually exclusive */
+
+/* is there a symbol table loaded? */
+#define MB_INFO_AOUT_SYMS		0x00000010
+/* is there an ELF section header table? */
+#define MB_INFO_ELF_SHDR		0x00000020
+
+/* is there a full memory map? */
+#define MB_INFO_MEM_MAP		0x00000040
+
+/* Is there drive info?  */
+#define MB_INFO_DRIVE_INFO		0x00000080
+
+/* Is there a config table?  */
+#define MB_INFO_CONFIG_TABLE	0x00000100
+
+/* Is there a boot loader name?  */
+#define MB_INFO_BOOT_LOADER_NAME	0x00000200
+
+/* Is there a APM table?  */
+#define MB_INFO_APM_TABLE		0x00000400
+
+/* Is there video information?  */
+#define MB_INFO_VIDEO_INFO		0x00000800
+
+struct multiboot_header {
+	/* Must be MB_MAGIC - see above.  */
+	uint32_t magic;
+
+	/* Feature flags.  */
+	uint32_t flags;
+
+	/* The above fields plus this one must equal 0 mod 2^32. */
+	uint32_t checksum;
+
+	/* These are only valid if MB_AOUT_KLUDGE is set.  */
+	uint32_t header_addr;
+	uint32_t load_addr;
+	uint32_t load_end_addr;
+	uint32_t bss_end_addr;
+	uint32_t entry_addr;
+
+	/* These are only valid if MB_VIDEO_MODE is set.  */
+	uint32_t mode_type;
+	uint32_t width;
+	uint32_t height;
+	uint32_t depth;
+};
+
+struct multiboot_info {
+	/* Multiboot info version number */
+	uint32_t flags;
+
+	/* Available memory from BIOS */
+	uint32_t mem_lower;
+	uint32_t mem_upper;
+
+	/* "root" partition */
+	uint32_t boot_device;
+
+	/* Kernel command line */
+	uint32_t cmdline;
+
+	/* Boot-Module list */
+	uint32_t mods_count;
+	uint32_t mods_addr;
+
+	uint32_t syms[4];
+
+	/* Memory Mapping buffer */
+	uint32_t mmap_length;
+	uint32_t mmap_addr;
+
+	/* Drive Info buffer */
+	uint32_t drives_length;
+	uint32_t drives_addr;
+
+	/* ROM configuration table */
+	uint32_t config_table;
+
+	/* Boot Loader Name */
+	uint32_t boot_loader_name;
+
+	/* APM table */
+	uint32_t apm_table;
+
+	/* Video */
+	uint32_t vbe_control_info;
+	uint32_t vbe_mode_info;
+	uint16_t vbe_mode;
+	uint16_t vbe_interface_seg;
+	uint16_t vbe_interface_off;
+	uint16_t vbe_interface_len;
+};
+
+struct multiboot_mmap_entry {
+	uint32_t size;
+	uint64_t addr;
+	uint64_t len;
+	uint32_t type;
+} __attribute__ ((packed));
+
+
+#define FIRMWARE_TYPE "coreboot"
+#define BOOTLOADER "elfboot"
+#define BOOTLOADER_VERSION "1.3"
+
+#endif
Index: i386/ldscript
===================================================================
--- i386/ldscript	(revision 2)
+++ i386/ldscript	(revision 7)
@@ -50,16 +50,28 @@
 
     /* Normal sections */
     .boot : { *(.boot) *(.boot.*) }
-    .text : { *(.text) *(.text.*) }
+    .text : { 
+		_text = .;
+		*(.text);
+		*(.text.*);
+		_etext = .;
+	}
     .rodata : {
+	_rodata = .;
 	. = ALIGN(4);
 	drivers_start = .;
 	*(.rodata.drivers)
 	drivers_end = .;
 	*(.rodata)
 	*(.rodata.*)
+	. = ALIGN(4);
+	_erodata = .;
     }
-    .data : { *(.data) *(.data.*) }
+    .data : {
+		_data = .;
+		*(.data) 
+		*(.data.*)
+		_edata = .; }
 
     .bss : {
 	*(.sbss)
@@ -76,6 +88,12 @@
 	. = ALIGN(16);
 	_eheap = .;
 
+	/* The ram segment
+ 	 * This is all address of the memory resident copy of filo.
+	 */
+	_ram_seg = _text; 
+	_eram_seg = _eheap;
+
 	_stack = .;
 	. += STACK_SIZE;
 	. = ALIGN(16);
Index: i386/Makefile.inc
===================================================================
--- i386/Makefile.inc	(revision 2)
+++ i386/Makefile.inc	(revision 7)
@@ -20,3 +20,4 @@
 TARGETS-$(CONFIG_LINUX_LOADER) += i386/linux_load.o
 TARGETS-$(CONFIG_WINCE_LOADER) += i386/wince_load.o
 TARGETS-$(CONFIG_ARTEC_BOOT) += i386/artecboot.o
+TARGETS-$(CONFIG_CBFS_BOOT) += i386/payload_boot.o
Index: i386/payload_boot.c
===================================================================
--- i386/payload_boot.c	(revision 0)
+++ i386/payload_boot.c	(revision 7)
@@ -0,0 +1,115 @@
+/*******************************************************************************
+ *
+ *	FILO Coreboot Payload loader, enables multiboot through custom header 
+ *
+ *	This program is free software; you can redistribute it and/or modify
+ *	it under the terms of the GNU General Public License as published by
+ *	the Free Software Foundation; either version 2 of the License, or
+ *	(at your option) any later version.
+ *
+ *	This program is distributed in the hope that it will be useful,
+ *	but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *	GNU General Public License for more details.
+ *
+ *	You should have received a copy of the GNU General Public License
+ *	along with this program; if not, write to the Free Software
+ *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ *
+ ******************************************************************************/
+#include <libpayload.h>
+#include <elf.h>
+#include <arch/elf.h>
+#include <elf_boot.h>
+#include <ipchecksum.h>
+#include <arch/multiboot.h>
+
+#ifndef CMD_LINE
+#define CMD_LINE ""
+#endif
+
+#define UPSZ(X) ((sizeof(X) + 3) &~3)
+
+static struct {
+	Elf_Bhdr hdr;
+	Elf_Nhdr ft_hdr;
+	unsigned char ft_desc[UPSZ(FIRMWARE_TYPE)];
+	Elf_Nhdr bl_hdr;
+	unsigned char bl_desc[UPSZ(BOOTLOADER)];
+	Elf_Nhdr blv_hdr;
+	unsigned char blv_desc[UPSZ(BOOTLOADER_VERSION)];
+	Elf_Nhdr cmd_hdr;
+	unsigned char cmd_desc[UPSZ(CMD_LINE)];
+} elf_boot_notes = {
+	.hdr = {
+		.b_signature = 0x0E1FB007,
+		.b_size = sizeof(elf_boot_notes),
+		.b_checksum = 0,
+		.b_records = 4,
+	},
+	.ft_hdr = {
+		.n_namesz = 0,
+		.n_descsz = sizeof(FIRMWARE_TYPE),
+		.n_type = EBN_FIRMWARE_TYPE,
+	},
+	.ft_desc = FIRMWARE_TYPE,
+	.bl_hdr = {
+		.n_namesz = 0,
+		.n_descsz = sizeof(BOOTLOADER),
+		.n_type = EBN_BOOTLOADER_NAME,
+	},
+	.bl_desc = BOOTLOADER,
+	.blv_hdr = {
+		.n_namesz = 0,
+		.n_descsz = sizeof(BOOTLOADER_VERSION),
+		.n_type = EBN_BOOTLOADER_VERSION,
+	},
+	.blv_desc = BOOTLOADER_VERSION,
+	.cmd_hdr = {
+		.n_namesz = 0,
+		.n_descsz = sizeof(CMD_LINE),
+		.n_type = EBN_COMMAND_LINE,
+	},
+	.cmd_desc = CMD_LINE,
+};
+
+struct multiboot_info *mbi;
+
+int elf_check_arch(Elf_ehdr *ehdr)
+{
+	return (
+		((ehdr->e_machine == EM_386) ||	(ehdr->e_machine == EM_486)) &&
+		(ehdr->e_ident[EI_CLASS] == ELFCLASS32) &&
+		(ehdr->e_ident[EI_DATA] == ELFDATA2LSB) 
+		);
+	
+}
+
+void jmp_to_elf_entry(void *entry)
+{
+	extern unsigned char _ram_seg, _eram_seg;
+	unsigned long lb_start, lb_size;
+	unsigned long adjust, adjusted_boot_notes;
+    struct sysinfo_t *info = &lib_sysinfo;
+
+	mbi = (struct multiboot_info *)info->mbtable;
+	
+	elf_boot_notes.hdr.b_checksum = 
+		ipchksum(&elf_boot_notes, sizeof(elf_boot_notes));
+
+	lb_start = (unsigned long) &_ram_seg;
+	lb_size = (unsigned long)(&_eram_seg - &_ram_seg);
+	
+	adjusted_boot_notes = (unsigned long) virt_to_phys(&elf_boot_notes);
+
+	
+	printf("entry    = 0x%08lx\n", (unsigned long)entry);
+	printf("lb_start = 0x%08lx\n", lb_start);
+	printf("lb_size  = 0x%08lx\n", lb_size);
+	printf("adjust   = 0x%08lx\n", adjust);
+	printf("     elf_boot_notes = 0x%08lx\n", (unsigned long)&elf_boot_notes);
+	printf("adjusted_boot_notes = 0x%08lx\n", adjusted_boot_notes);
+	
+	start_elf(entry, adjusted_boot_notes);
+	
+}
Index: Config.in
===================================================================
--- Config.in	(revision 2)
+++ Config.in	(revision 7)
@@ -213,6 +213,10 @@
 	bool "Squash filesystem"
 	default n
 
+config FSYS_CBFS
+	bool "CBFS ROM Image"
+	default y
+
 endmenu
 
 menu "Loaders"
@@ -238,6 +242,12 @@
 	help
 	  Artecboot Loader Support
 
+config CBFS_BOOT
+	bool "CBFS Payload Loader"
+	default n
+	depends on FSYS_CBFS
+	help
+		Coreboot CBFS SELF Payload Support
 endmenu
 
 menu "Debugging & Experimental"
@@ -338,6 +348,11 @@
 	depends on ARTEC_BOOT
 	default n
 
+config DEBUG_CBFSBOOT
+	bool "DEBUG_CBFSBOOT"
+	depends on CBFS_BOOT
+	default n
+	
 config DEVELOPER_TOOLS
 	bool "Developer Tools"
 	depends on USE_GRUB
